// Copyright (c) 2014, The Shelf REST project authors.
// Please see the AUTHORS file for details.
// All rights reserved. Use of this source code is governed by
// a BSD 2-Clause License that can be found in the LICENSE file.

library shelf_rest.resource.test;


import 'package:shelf_rest/src/resource.dart';
import 'package:shelf/shelf.dart';
import 'package:shelf_route/shelf_route.dart';
import 'package:unittest/unittest.dart';
import 'package:mock/mock.dart';

class MockRouter extends Mock implements Router {
  noSuchMethod(Invocation invocation) => super.noSuchMethod(invocation);
}

main() {
  group('createRoutes', () {
    MockRouter router;
    MockRouter childRouter;
    MockRouter grandChildRouter;

    testHandler(Request request) => new Response.ok('all good');

    setUp(() {
      router = new MockRouter();
      childRouter = new MockRouter();
      grandChildRouter = new MockRouter();
      router.when(callsTo('addChildRouter')).alwaysReturn(childRouter);
      childRouter.when(callsTo('addChildRouter')).alwaysReturn(grandChildRouter);
    });

    skip_test('should throw ArgumentError if null path parameter', () {
      expect(() => (new TestResource()
        ..putHandler = testHandler)
        .createRoutes(router),
          throwsArgumentError);
    });

    test('should NOT create routes if NO handler present', () {
      new TestResource('fred')..createRoutes(router);
      router.verifyZeroInteractions();
    });

    test('should create search route if handler present', () {
      new TestResource('fred')
        ..searchHandler = testHandler
        ..createRoutes(router);
      router.calls('get', '/', equals(testHandler)).verify(happenedOnce);
      childRouter.verifyZeroInteractions();
    });

    test('should include query parameters in search route if present', () {
      new TestResource('fred')
        ..searchHandler = testHandler
        ..searchQueryParameters = ['foo', 'bar']
        ..createRoutes(router);
      router.calls('get', '{?foo,bar}', equals(testHandler)).verify(happenedOnce);
      childRouter.verifyZeroInteractions();
    });

    test('should create post route if handler present', () {
      new TestResource('fred')
        ..postHandler = testHandler
        ..createRoutes(router);
      router.calls('post', '/', equals(testHandler)).verify(happenedOnce);
      childRouter.verifyZeroInteractions();
    });

    test('should create find route if handler present', () {
      new TestResource('fred')
        ..findHandler = testHandler
        ..createRoutes(router);
      router.calls('addChildRouter','/{fred}').verify(happenedOnce);
      childRouter.calls('get', '/', equals(testHandler)).verify(happenedOnce);
    });

    test('should create put route if handler present', () {
      new TestResource('fred')
        ..putHandler = testHandler
        ..createRoutes(router);
      router.calls('addChildRouter','/{fred}').verify(happenedOnce);
      childRouter.calls('put', '/', equals(testHandler)).verify(happenedOnce);
    });

    test('should create delete route if handler present', () {
      new TestResource('fred')
        ..deleteHandler = testHandler
        ..createRoutes(router);
      router.calls('addChildRouter','/{fred}').verify(happenedOnce);
      childRouter.calls('delete', '/', equals(testHandler)).verify(happenedOnce);
    });

    test('should add routes in child resource', () {
      final childResource = new TestResource('barney')
        ..searchHandler = testHandler;

      new TestResource('fred')
        ..childResources = { '/barney': childResource }
        ..createRoutes(router);

      router.calls('addChildRouter','/{fred}').verify(happenedOnce);
      childRouter.calls('addChildRouter','/barney').verify(happenedOnce);
      grandChildRouter.calls('get', '/', equals(testHandler)).verify(happenedOnce);
    });
  });
}


class TestResource extends Resource {
  @override
  String pathParameterName;

  Handler searchHandler;
  Handler findHandler;
  Handler postHandler;
  Handler putHandler;
  Handler deleteHandler;
  Iterable<String> searchQueryParameters = const [];

  TestResource([this.pathParameterName]);

  Map<dynamic, Resource> childResources = const {};

}

